/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.service;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.fortiss.tooling.base.model.element.IHierarchicElement;

/**
 * {@link IModelQualityProvider} to count the ratio of filled out comments of
 * INamedCommentedElements.
 * 
 * @author blaschke
 */
public class HierarchicElementSizeProvider implements IModelQualityProvider<IHierarchicElement> {
	/**
	 * returns an array of integers or <String,Double> Map
	 * Number of children of the hierarchical Element (on this level)
	 * Number of ports on the hierarchical Element (on this level)
	 * Number of channels in the hierarchical element (on this level)
	 * Number of totalPorts in whole tree
	 * Number of totalElements in whole tree
	 * Average SubTree Size
	 */
	/* */
	private final String rootName;
	/**  */
	Map<String, Double> sizes = new HashMap<String, Double>();

	/** */
	public HierarchicElementSizeProvider(String rootName) {
		super();
		this.rootName = rootName;
		// TODO Auto-generated constructor stub
	}

	/** */
	public String getName() {
		return this.rootName;
	}

	/** {@inheritDoc} */
	@Override
	public void apply(IHierarchicElement ele) {
		Map<String, Double> sizes = new HashMap<String, Double>();
		sizes.put("numberOfPorts", (double)ele.getConnectors().size());
		sizes.put("numberOfChildren", (double)ele.getContainedElements().size());
		sizes.put("numberOfChannels", (double)ele.getConnections().size());

		// depth metrics
		Set<ICounter> depthMetrics = new HashSet<HierarchicElementSizeProvider.ICounter>();
		PortCounter pc = new PortCounter();
		ElementCounter ec = new ElementCounter();
		depthMetrics.add(new PortCounter());
		depthMetrics.add(new ElementCounter());
		// traverse whole tree
		traverse(ele, depthMetrics);

		// put in entries for subtrees
		sizes.put("numberOfTotalPorts:", (double)pc.getCounter());
		sizes.put("numberOfTotalElements:", (double)ec.getCounter());

		for(IHierarchicElement child : ele.getContainedElements()) {
			apply(child);
		}
		this.sizes = sizes;
	}

	/**
	 * @param currentElement
	 * @param consumers
	 */
	public void traverse(IHierarchicElement currentElement, Set<ICounter> consumers) {
		consumers.forEach(c -> c.apply(currentElement));
		for(IHierarchicElement child : currentElement.getContainedElements()) {
			traverse(child, consumers);
		}
	}

	/** */
	public interface ICounter {
		/**
		 * @param ele
		 */
		public void apply(IHierarchicElement ele);

		/** */
		public int getCounter();
	}

	/** */
	private class PortCounter implements ICounter {
		/** Returns port number. */
		int counter = 0;

		/** {@inheritDoc} */
		@Override
		public int getCounter() {
			return counter;
		}

		/** */
		@Override
		public void apply(IHierarchicElement element) {
			counter = getCounter() + element.getConnectors().size();
		}
	}

	/** */
	private class ElementCounter implements ICounter {
		/** Returns port number. */
		int counter = 0;

		/** {@inheritDoc} */
		@Override
		public int getCounter() {
			return counter;
		}

		/** */
		@Override
		public void apply(IHierarchicElement element) {
			counter += 1;
		}
	}
}

/*
 * public abstract class MultiMethodDemo implements IElementConsumer {
 * 
 * @Override
 * public void apply(IHierarchicElement element) {
 * if(element instanceof INamedCommentedElement) {
 * consume((INamedCommentedElement)element);
 * } else {
 * consume(element);
 * }
 * }
 * 
 * public abstract void consume(IHierarchicElement element);
 * 
 * public abstract void consume(INamedCommentedElement element);
 * 
 * }
 */

// public class CloneDetection implements IElementConsumer {
//
// private Map<Integer, List<IHierarchicElement>> map = new HashMap<>();
//
// @Override
// public void printResult() {
// System.out.println("Stuff");
// }
//
// @Override
// public void apply(IHierarchicElement element) {
// int hash = 1;
// for(IConnector connector : element.getConnectors()) {
// EList<IModelElementSpecification> specifications = connector.getSpecifications();
// //if(!specifications.isEmpty() &&
// // connector.getSpecifications().get(0) instanceof PortSpecification) {
//
// // PortSpecification port =
// // (PortSpecification)connector.getSpecifications().get(0);
// // hash = 31 * hash + (connector instanceof InputPort ? 1 : 2) *
// // port.getType().getClass().hashCode();
// } else {
// // Clone Detection for State diagrams not supported
// return;
// }
// }map.putIfAbsent(hash,new ArrayList<>());
//
// List<IHierarchicElement> elements = map.get(hash);
//
// for(
// IHierarchicElement element_b:elements)
// {
//
// List<String> element_a_input = new ArrayList<String>();
// List<String> element_b_input = new ArrayList<String>();
// List<String> element_a_output = new ArrayList<String>();
// List<String> element_b_output = new ArrayList<String>();
//
// for(IConnector connector : element.getConnectors()) {
// PortSpecification port = (PortSpecification)connector.getSpecifications().get(0);
// String type = port.getType().getClass().getSimpleName();
// if(connector instanceof InputPort) {
// element_a_input.add(type);
// } else {
// element_a_output.add(type);
// }
// }
//
// for(IConnector connector : element_b.getConnectors()) {
// PortSpecification port = (PortSpecification)connector.getSpecifications().get(0);
// String type = port.getType().getClass().getSimpleName();
// if(connector instanceof InputPort) {
// element_b_input.add(type);
// } else {
// element_b_output.add(type);
// }
// }
//
// boolean equals = true;
//
// for(String type_a_input : element_a_input) {
// if(!element_b_input.remove(type_a_input)) {
// equals = false;
// break;
// }
//
// }
//
// for(String type_a_output : element_a_output) {
// if(!element_b_output.remove(type_a_output)) {
// equals = false;
// break;
// }
//
// }
// if(equals) {
// System.out
// .println("Found duplicate: " + ((INamedCommentedElement)element).getName() +
// ", " + ((INamedCommentedElement)element_b).getName());
// break;
// } else {
// System.out.println("Same but not duplicate");
// }
// }elements.add(element);
// }

/*
 * public class CommentCompleteness extends MultiMethodDemo {
 * private int total_amount;
 * private int with_comments;
 * 
 * public CommentCompleteness() {
 * total_amount = 0;
 * with_comments = 0;
 * }
 * 
 * @Override
 * public void consume(IHierarchicElement element) {
 * total_amount++;
 * }
 * 
 * @Override
 * public void consume(INamedCommentedElement element) {
 * if(element.getComment() != "") {
 * with_comments++;
 * }
 * total_amount++;
 * }
 * }
 */
/*
 * public class Tuple<T1, T2> {
 * private T1 element1;
 * private T2 element2;
 * 
 * public Tuple(T1 element1, T2 element2) {
 * this.element1 = element1;
 * this.element2 = element2;
 * }
 * 
 * public T1 getElement1() {
 * return element1;
 * }
 * 
 * public T2 getElement2() {
 * return element2;
 * }
 * }
 */
/*
 * public void printShare(HierarchicElementBase element) {
 * 
 * Set<Tuple<String, Integer>> counts = new HashSet<>();
 * this.apply(element);
 * counts.add(new Tuple<>("[Self]" + element.getName(), count));
 * count = 0;
 * 
 * for(IHierarchicElement child : element.getContainedElements()) {
 * 
 * HierarchicElementBase real_child = (HierarchicElementBase)child;
 * 
 * recursiveDescent(real_child, this);
 * 
 * counts.add(new Tuple<>(real_child.getName(), count));
 * count = 0;
 * }
 * 
 * double total = counts.stream().mapToInt(t -> t.getElement2()).sum();
 * 
 * System.out.println("[" + this.getClass().getSimpleName() + "] shares:");
 * for(Tuple<String, Integer> tuple : counts) {
 * System.out.println("\t " + String.format("%05.2f%%", tuple.element2 / total * 100) +
 * ", " + tuple.element2 + ": " + tuple.element1);
 * }
 * }
 */

/*
 * protected void recursiveDescent(HierarchicElementBase element, CounterConsumer consumer) {
 * consumer.apply(element);
 * 
 * for(IHierarchicElement child : element.getContainedElements()) {
 * HierarchicElementBase real_child = (HierarchicElementBase)child;
 * recursiveDescent(real_child, consumer);
 * }
 * }
 */

/*
 * public class ConnectionCounter {
 * public void apply(IHierarchicElement element) {
 * count += element.getConnections().size();
 * }
 * 
 * }
 * 
 * private void extractComment(IHierarchicElement c) {
 * strings.add(((INamedCommentedElement)c).getComment());
 * }
 * }
 * }
 */
