/*-------------------------------------------------------------------------+
| Copyright 2013 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.quality.service;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstChildWithType;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.project.model.FileProject;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IMigrationService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;

/**
 * This class implements the {@link IMigrationService} interface.
 * 
 * @author blaschke
 */
public class ModelQualityService extends EObjectAwareServiceBase<IModelQualityProvider<?>>
		implements IIntrospectiveKernelService, IModelQualityService {

	/** The singleton instance. */
	private static final ModelQualityService INSTANCE = new ModelQualityService();

	/** Returns singleton instance of the service. */
	public static ModelQualityService getInstance() {
		return INSTANCE;
	}

	/** The connector extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.ext.quality.metricService";

	/** The connector configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "metricService";

	/** The connector attribute name. */
	private static final String ATTRIBUTE_NAME = "metricService";

	/** */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** Registers the migration provider with the service. */
	@Override
	public void registerMetricProvider(IModelQualityProvider<?> provider,
			Class<?> modelElementClass) {
		addHandler(modelElementClass, provider);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service allows the tracking of metrical development in modeling." +
				"\n the service outputs the current metric analysis of the model and stores" +
				"\n the data within logs for later history analysis." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/** Get all suitable {@link IModelQualityProvider} for the given input. */
	private List<IModelQualityProvider<?>> getAllMetricProviders(EObject input) {
		List<IModelQualityProvider<?>> providers = new ArrayList<>();
		for(Entry<Class<?>, List<IModelQualityProvider<?>>> migEntry : handlerMap.entrySet()) {
			if(getFirstChildWithType(input, migEntry.getKey()) != null) {
				providers.addAll(migEntry.getValue());
			}
		}
		System.out.println("These are the providers:" + providers);
		return providers;
	}

	/**
	 * @param fp
	 * @return
	 */
	public List<IModelQualityProvider<? extends EObject>> performMetricAnalysis(FileProject fp) {
		List<IModelQualityProvider<? extends EObject>> result =
				new LinkedList<IModelQualityProvider<? extends EObject>>();
		for(IProjectRootElement ele : fp.getRootElements()) {
			if(ele instanceof IHierarchicElement) {
				HierarchicElementSizeProvider hes =
						new HierarchicElementSizeProvider(ele.getName());
				hes.apply((IHierarchicElement)ele);
				System.out.println("checked in hierarchic metrics for: " + hes.getName());
				result.add(hes);
			}
		}

		// List<IModelQualityProvider<?>> handlers = getRegisteredHandlers(modelElement.getClass());
		// Map<String, Double> qualityData = new HashMap<String, Double>();

		// getRegisteredHandlers(IHierarchicElement.class);
		// List<IModelQualityProvider<?>> hierarchicHandlers2 = getAllMetricProviders(fp);
		// if(handlers == null) {
		// return null;
		// }

		// for(IModelQualityProvider<?> qualityProv : handlers) {
		// List<? extends IModelQualityProvider<? extends EObject>> collectedProviders =
		// qualityProv.apply(modelElement);
		// if(metrics != null) {
		// metricsList.addAll(metrics);
		// }
		// }
		return result;
	}

	/** */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** */
	@Override
	protected String getHandlerClassAttribute() {
		return ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Metric Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	@SuppressWarnings("javadoc")
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void metricExtractionToCSV(Map<String, Double> data) {
		MetricCSVWriter writer = new MetricCSVWriter();
		try {
			writer.openFile();
			writer.writeData(data);
		} catch(IOException e) {
			e.printStackTrace();
		} finally {
			try {
				writer.closeFile();
			} catch(IOException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * 
	 * @author blaschke
	 */
	public static class MetricCSVWriter {
		/**  */
		private static final String FILE_PATH = "data_ultra_cool.csv";
		/**  */
		private BufferedWriter writer;

		/**
		 * @throws IOException
		 */
		public void openFile() throws IOException {
			writer = new BufferedWriter(new FileWriter(FILE_PATH, true));
		}

		/**
		 * @param data
		 */
		public void writeData(Map<String, Double> data) {
			try {
				String timestamp = getCurrentTimestamp();
				writer.write(timestamp + ": ");
				for(Map.Entry<String, Double> entry : data.entrySet()) {
					writer.write(entry.getKey() + ":" + entry.getValue() + ",");
					writer.newLine();
				}
				writer.flush();
			} catch(IOException e) {
				e.printStackTrace();
			}
		}

		/**
		 * @throws IOException
		 */
		public void closeFile() throws IOException {
			if(writer != null) {
				writer.close();
			}
		}

		/** */
		private String getCurrentTimestamp() {
			SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
			Date now = new Date();
			return dateFormat.format(now);
		}
	}
}
